import sys
from PyQt5 import QtCore, QtWidgets, uic
from PyQt5.QtCore import Qt
from loan import Loan, LoanError

welcome_msg = "Emprunt à intérêts composés et remboursement mensuel " \
              "constant ; voir les bulles d'aide des boutons pour " \
              "les détails"

class LoanWindow(QtWidgets.QMainWindow):
    def __init__(self, capital, year_rate, time, prime):

        QtWidgets.QMainWindow.__init__(self)

        # La vue : le fichier qtloan.ui, créé par le designer de qt (fourni
        # avec la distribution) contient la description de l'interface
        # graphique. C'est là que sont définis tous les objets self.button_...,
        # self.input_..., etc.
        uic.loadUi("qtloan.ui", self)
        self.setAttribute(Qt.WA_DeleteOnClose, True)
        self.status_bar = self.statusBar()
        self.label_message.setText(welcome_msg)

        # Remplissage du menu de choix mois/année
        for entry in ("mois", "années"):
            self.input_month_year.addItem(entry)
        self._display_time = 'month'

        # Le modèle : contient les données à afficher et les algos de calcul
        self._loan = Loan(capital=capital, year_rate=year_rate,
                         time=time, prime=prime)

        # Tranfert du focus du label au champ d'édition
        self.label_capital.setFocusProxy(self.input_capital)
        self.label_year_rate.setFocusProxy(self.input_year_rate)
        self.label_time.setFocusProxy(self.input_time)
        self.label_prime.setFocusProxy(self.input_prime)

        # Le contrôleur : on associe les interactions utilisateurs à des
        # méthodes
        self.button_prime.clicked.connect(self.show_prime)
        self.button_year_rate.clicked.connect(self.show_year_rate)
        self.button_time.clicked.connect(self.show_time)
        self.button_plan.clicked.connect(self.show_plan)
        self.button_close.clicked.connect(self.close)

        self.input_capital.returnPressed.connect(self.show_prime)
        self.input_year_rate.returnPressed.connect(self.show_prime)
        self.input_time.returnPressed.connect(self.show_prime)
        self.input_prime.returnPressed.connect(self.show_year_rate)
        self.input_month_year.currentIndexChanged.connect(self.set_month_year)

        self.update()

    def message(self, msg, reason=None):
        """Affiche msg"""
        self.status_bar.showMessage(msg)
        if reason is not None:
            entry = getattr(self, 'input_' + reason, None)
            if entry is not None:
                entry.setFocus()
                entry.selectAll()

    def check_entries(self, ignore="prime"):
        """
        Lit chacune des données fournies par l'utilisateur (sauf possiblement
        celle spécifiée par ignore qui contient la donnée qui va être
        recalculée en fonction des autres) et met à jour le modèle en
        conséquence
        """
        for entry_name in ('capital', 'year_rate', 'prime'):
            if entry_name == ignore:
                continue

            # On récupère le contenu de l'entrée input_<entry_name>
            input = getattr(self, 'input_' + entry_name).text()
            if input:
                # Mise à jour de la donnée entry_name de l'objet modèle ; si la
                # donnée utilisateur est incorrecte ceci lèvera l'exception
                # LoanError
                getattr(self._loan, entry_name)(input)

        # La durée peut être exprimée en mois ou année selon le choix
        # utilisateur
        if ignore != 'time' and self.input_time.text():
            try:
                time = int(self.input_time.text())
                if self._display_time == 'year':
                    time *= 12
                self._loan.time(time)
            except ValueError:
                raise LoanError("La durée doit être un nombre entier positif",
                                'time')

    def update(self, msg=None):
        """
        Met à jour l'affichage en fonction des données contenues dans le
        modèle ; msg contient un message à afficher.
        """
        for (entry_name, format) in (('capital', "%.2f"),
                                     ('prime', "%.2f"),
                                     ('year_rate', "%.2f")):
            entry = getattr(self, 'input_' + entry_name)
            value = getattr(self._loan, entry_name)()
            if value:
                entry.setText(format % value)

        # La durée peut être affichée en mois ou en année selon le choix
        # utilisateur
        time = self._loan.time()
        if time is not None:
            if self._display_time == 'year':
                time //= 12
            self.input_time.setText(str(time))

        if self._loan.total():
            self.input_total.setText("%.2f" % self._loan.total())
            self.input_cost.setText(
                "%.2f" % (self._loan.total() - self._loan.capital()))

        self.message(msg);

    def show_prime(self):
        """
        Action associée au bouton "Mensualité"
        """
        try:
            self.check_entries('prime')
            self._loan.compute_prime()
            self.update("Calcul de la prime mensuelle")

        except LoanError as msg:
            self.message(str(msg), msg.reason)

    def set_month_year(self, index):
        """
        Action associée au choix "mois/année" ; on met à jour l'entrée time et
        la valeur de self._display_time
        """

        new_display = 'year' if self.input_month_year.currentIndex() == 1 \
                      else 'month'

        if new_display == self._display_time:
            return

        self._display_time = new_display
        time = self.input_time.text()
        if time:
            try:
                time = int(time)
                if new_display == 'month':
                    time *= 12
                else:
                    time //= 12
                self.input_time.setText(str(time))
                self.message(
                    "Durée en %s" % ("années" if new_display == 'year'
                                     else "mois"), 'time')

            except ValueError:
                self.message("La durée doit être un entier positif", 'time')
        else:
            self.message("Entrer une durée", 'time')

    def show_time(self):
        """
        Action associée au bouton "Durée"
        """
        try:
            self.check_entries('time')
            self._loan.compute_time()
            self.update("Calcul du nombre de mois de remboursement")

        except LoanError as msg:
            self.message(str(msg), msg.reason)

    def show_year_rate(self):
        """
        Action associée au bouton "Taux annuel"
        """
        try:
            self.check_entries('year_rate')
            self._loan.compute_rate()
            self.update("Calcul du taux annuel")

        except LoanError as msg:
            self.message(str(msg), msg.reason)

    def show_plan(self):
        """
        Action associée au bouton "Plan de remboursement"
        """
        try:
            self.check_entries()
            plan = self._loan.plan()
            self.update("Plan de remboursement")

        except LoanError as msg:
            self.message(str(msg), msg.reason)
            return

        self.plan_window = PlanWindow(self, plan)
        self.plan_window.show()

    def export(self):
        """
        Export CSV du plan de remboursement
        """
        try:
            self.check_entries()
            plan = self._loan.plan()
            self.update("Export CSV")

        except LoanError as msg:
            self.message(str(msg), msg.reason)
            return

        export_file = str(QtWidgets.QFileDialog.getSaveFileName(
            self,
            "Exporter un fichier cvs",
            "",
            "cvs file (*.csv)")[0])

        if not export_file:
            self.message("Export annulé")
            return

        try:
            self._loan.export(export_file, plan)
            self.message("Fichier %s sauvé" % export_file)

        except LoanError as msg:
            self.message(str(msg), msg.reason)

class PlanWindow(QtWidgets.QMainWindow):
    """
    La fenêtre affichant le plan de remboursement
    """

    def __init__(self, parent, plan):
        QtWidgets.QMainWindow.__init__(self, parent)
        # Le fichier qtloan_plan.ui créé avec le designer de qt contient la
        # description de l'interface graphique
        uic.loadUi("qtloan_plan.ui", self)

        table = self.tableView
        table.setModel(PlanModel(plan))

        self.h_header = table.horizontalHeader()
        self.v_header = table.verticalHeader()

        font = table.font()
        font.setBold(True)
        self.h_header.setFont(font)
        self.v_header.setFont(font)

        self.button_export.clicked.connect(parent.export)
        self.setAttribute(Qt.WA_DeleteOnClose, True)

class PlanModel(QtCore.QAbstractTableModel):
    """
    Implémentation du plan de remboursement : le modèle affiché par la vue
    PlanWindow. Cf la doc qt de QAbstractItemModel pour détails
    """

    def __init__(self, plan):
        QtCore.QAbstractTableModel.__init__(self)
        self.plan = plan
        self.header = (u"Capital", u"Intérêt", u"Total remboursé",
                       u"dont intérêt")
        self.tooltips = (u"Montant du capital restant dû",
                         u"Montant de l'intérêt du mois",
                         u"Total remboursé, i.e. numéro du mois x montant de la prime",
                         u"Total des intérêts remboursés")

    def columnCount(self, index):
        return 4

    def rowCount(self, index):
        return len(self.plan)

    def data(self, index, role):
        """
        Contenu d'une cellule
        """
        if role == Qt.DisplayRole:
            data = self.plan[index.row()]
            return "%.2f" % data[index.column()+1]
        elif role == Qt.ToolTipRole:
            return self.tooltips[index.column()]

        else:
            return None

    def headerData(self, index, orientation, role):
        """
        Contenu d'une cellule d'en-tête
        """
        if orientation == Qt.Horizontal:
            if role == Qt.DisplayRole:
                return self.header[index]
            elif role == Qt.ToolTipRole:
                return self.tooltips[index]
            else:
                return None

        else:
            if role == Qt.DisplayRole:
                return "Mois %2d" % index
            else:
                return None

# On peut exécuter directement le présent script mais il vaut mieux passer par
# ./qtloan qui permet de passer des options

if __name__ == '__main__':
    app = QtWidgets.QApplication(sys.argv)
    win = LoanWindow(sys.argv)
    win.show()
    sys.exit(app.exec_())
